# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import scipy.io as sio
from scipy.linalg import hankel
from ..tools.tools import nextpow2,flat_eq
def calculate_bicoherence(y, nfft=None, wind=None, nsamp=None, overlap=None):
    """
    Direct (FD) method for estimating bicoherence.

    Args:
        y: data vector or time-series
        nfft: FFT length [default = power of two greater than nsamp]
            actual size used is the power of two greater than ``nsamp``
        wind: time-domain window applied to each data segment
            should be of length ``segsamp`` (see below); otherwise,
            the default Hanning window is used
        nsamp: samples per segment [default: such that there are 8 segments]
            if ``y`` is a matrix, ``nsamp`` is set to the number of rows
        overlap: percentage overlap, allowed range [0, 99] [default = 50]
            if ``y`` is a matrix, overlap is set to 0

    Returns:
        bic: estimated bicoherence, an nfft x nfft array, with origin
            at the center, and axes pointing down and to the right
        waxis: vector of frequencies associated with the rows and columns
            of ``bic``; sampling frequency is assumed to be 1
    """
    # Parameter checks
    (ly, nrecs) = y.shape
    if ly == 1:
        y = y.reshape(1, -1)
        ly = nrecs
        nrecs = 1

    if not nfft: nfft = 128
    if not overlap: overlap = 50
    overlap = max(0, min(overlap, 99))
    if nrecs > 1: overlap = 0
    if not nsamp: nsamp = 0
    if nrecs > 1: nsamp = ly

    if nrecs ==1 and nsamp <= 0:
        nsamp = np.fix(ly / (8 - 7 * overlap / 100))
    if nfft < nsamp:
        nfft = nextpow2(nsamp)

    overlap = np.fix(nsamp * overlap / 100)
    nadvance = nsamp - overlap
    nrecs = np.fix((ly * nrecs - overlap) / nadvance)

    if not wind:
        wind = np.hanning(nsamp)

    try:
        (rw, cw) = wind.shape
    except ValueError:
        (rw,) = wind.shape
        cw = 1

    if min(rw, cw) == 1 or max(rw, cw) == nsamp:
        print("Segment size is " + str(nsamp))
        print("Wind array is " + str(rw) + " by " + str(cw))
        print("Using default Hanning window")
        wind = np.hanning(nsamp)

    wind = wind.reshape(1, -1)

    # Accumulate triple products

    bic = np.zeros([nfft, nfft])
    Pyy = np.zeros([nfft, 1])

    mask = hankel(np.arange(nfft), np.array([nfft - 1] + list(range(nfft - 1))))
    Yf12 = np.zeros([nfft, nfft])
    ind = np.arange(nsamp,dtype=int)
    y = y.ravel(order='F')


    for k in np.arange(nrecs):
        ys = y[ind]
        ys = (ys.reshape(1, -1) - np.mean(ys)) * wind

        Yf = np.fft.fft(ys, nfft) / nsamp
        CYf = np.conjugate(Yf)
        Pyy = Pyy + flat_eq(Pyy, (Yf * CYf))

        Yf12 = flat_eq(Yf12, CYf.ravel(order='F')[mask])

        bic = bic + ((Yf * np.transpose(Yf)) * Yf12)
        ind = ind + int(nadvance)

    bic = bic / nrecs
    Pyy = Pyy / nrecs
    mask = flat_eq(mask, Pyy.ravel(order='F')[mask])
    bic = abs(bic) ** 2 / ((Pyy * np.transpose(Pyy)) * mask)
    bic = np.fft.fftshift(bic)

    if nfft % 2 == 0:
        waxis = np.transpose(np.arange(-1 * nfft / 2, nfft / 2)) / nfft
    else:
        waxis = np.transpose(np.arange(-1 * (nfft - 1) / 2, (nfft - 1) / 2 + 1)) / nfft

    cont = plt.contourf(waxis, waxis, bic, 4, cmap=plt.cm.Spectral_r)#100
    plt.colorbar(cont)
    plt.title('Bicoherence estimated via the direct (FFT) method')
    plt.xlabel('f1')
    plt.ylabel('f2')

    colmax, row = bic.max(0), bic.argmax(0)
    maxval, col = colmax.max(0), colmax.argmax(0)
    row=row[col]
    print('Max: bic(' + str(waxis[col]) + ',' + str(waxis[col]) + ') = ' + str(maxval))
    plt.show()

    return bic, waxis


def test():
    from ..tools.tools import here
    qpc = sio.loadmat(here(__file__) + '/demo/qpc.mat')
    dbic = calculate_bicoherence(qpc['zmat'])


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import scipy.io as sio
from scipy.linalg import hankel
from scipy.signal import convolve2d

from ..tools.tools import nextpow2, flat_eq, here


def bispectrumd(y, nfft=None, wind=None, nsamp=None, overlap=None):
    """
    Direct (FD) method for estimating the bispectrum.

    Parameters
    ----------
    y : array-like
        Data vector or time-series.
    nfft : int, optional
        FFT length. Default is the next power of two greater than the
        number of samples.
    wind : int or array-like, optional
        Window specification for frequency-domain smoothing.

        If ``wind`` is a scalar, it specifies the length of the side
        of the square for the Rao-Gabr optimal window (default is 5).

        If ``wind`` is a vector, a 2D window is calculated as::

            w2(i, j) = wind(i) * wind(j) * wind(i + j)

        If ``wind`` is a matrix, it specifies the 2-D filter directly.
    nsamp : int, optional
        Number of samples per segment. Default is chosen such that
        there are 8 segments.

        If ``y`` is a matrix, ``nsamp`` is set to the number of rows.
    overlap : float, optional
        Percentage overlap between segments. Default is 50.

        If ``y`` is a matrix, overlap is set to 0.

    Returns
    -------
    Bspec : ndarray
        Estimated bispectrum, an ``nfft × nfft`` array, with the origin
        at the center and axes pointing down and to the right.
    waxis : ndarray
        Vector of frequencies associated with the rows and columns of
        ``Bspec``. The sampling frequency is assumed to be 1.
    """
    (ly, nrecs) = y.shape
    if ly == 1:
        y = y.shape(1, -1)
        ly = nrecs
        nrecs = 1

    if not nfft: nfft = 128
    if not overlap: overlap = 50
    overlap = min(99, max(overlap, 0))
    if nrecs > 1: overlap = 0
    if not nsamp: nsamp = 0
    if nrecs > 1: nsamp = ly
    if nrecs == 1 and nsamp <= 0:
        nsamp = np.fix(ly / (8 - 7 * overlap / 100))
    if nfft < nsamp:
        nfft = 2 ** nextpow2(nsamp)
    overlap = np.fix(nsamp * overlap / 100)
    nadvance = nsamp - overlap
    nrecs = np.fix((ly * nrecs - overlap) / nadvance)

    if not wind: wind = 5

    m = n = 0
    try:
        (m, n) = wind.shape
    except ValueError:
        (m,) = wind.shape
        n = 1
    except AttributeError:
        m = n = 1

    window = wind
    # scalar: wind is size of Rao-Gabr window
    if max(m, n) == 1:
        winsize = wind
        if winsize < 0: winsize = 5  # the window size L
        winsize = winsize - (winsize % 2) + 1  # make it odd
        if winsize > 1:
            mwind = np.fix(nfft / winsize)  # the scale parameter M
            lby2 = (winsize - 1) / 2

            theta = np.array([np.arange(-1 * lby2, lby2 + 1)])  # force a 2D array
            opwind = np.ones([winsize, 1]) * (theta ** 2)  # w(m,n) = m**2
            opwind = opwind + opwind.transpose() + (np.transpose(theta) * theta)  # m**2 + n**2 + mn
            opwind = 1 - ((2 * mwind / nfft) ** 2) * opwind
            Hex = np.ones([winsize, 1]) * theta
            Hex = abs(Hex) + abs(np.transpose(Hex)) + abs(Hex + np.transpose(Hex))
            Hex = (Hex < winsize)
            opwind = opwind * Hex
            opwind = opwind * (4 * mwind ** 2) / (7 * np.pi ** 2)
        else:
            opwind = 1

    # 1-D window passed: convert to 2-D
    elif min(m, n) == 1:
        window = window.reshape(1, -1)

        if np.any(np.imag(window)) != 0:
            print("1-D window has imaginary components: window ignored")
            window = 1

        if np.any(window) < 0:
            print("1-D window has negative components: window ignored")
            window = 1

        lwind = np.size(window)
        w = window.ravel(order='F')
        # the full symmetric 1-D
        windf = np.array(w[range(lwind - 1, 0, -1) + [window]])
        window = np.array([window], np.zeros([lwind - 1, 1]))
        # w(m)w(n)w(m+n)
        opwind = (windf * np.transpose(windf)) * hankel(np.flipud(window), window)
        winsize = np.size(window)

    # 2-D window passed: use directly
    else:
        winsize = m

        if m != n:
            print("2-D window is not square: window ignored")
            window = 1
            winsize = m

        if m % 2 == 0:
            print("2-D window does not have odd length: window ignored")
            window = 1
            winsize = m

        opwind = window

    # accumulate triple products
    Bspec = np.zeros([nfft, nfft])  # the hankel mask (faster)
    mask = hankel(np.arange(nfft), np.array([nfft - 1] + list(range(nfft - 1))))
    locseg = np.arange(nsamp).transpose()
    y = y.ravel(order='F')

    for krec in np.arange(nrecs):
        xseg = y[locseg].reshape(1, -1)
        Xf = np.fft.fft(xseg - np.mean(xseg), nfft) / nsamp
        CXf = np.conjugate(Xf).ravel(order='F')
        Bspec = Bspec + \
                flat_eq(Bspec, (Xf * np.transpose(Xf)) * CXf[mask].reshape(nfft, nfft))
        locseg = locseg + int(nadvance)

    Bspec = np.fft.fftshift(Bspec) / nrecs

    # frequency-domain smoothing
    if winsize > 1:
        lby2 = int((winsize - 1) / 2)
        Bspec = convolve2d(Bspec, opwind)
        Bspec = Bspec[range(lby2 + 1, lby2 + nfft + 1), :][:, np.arange(lby2 + 1, lby2 + nfft + 1)]

    if nfft % 2 == 0:
        waxis = np.transpose(np.arange(-1 * nfft / 2, nfft / 2)) / nfft
    else:
        waxis = np.transpose(np.arange(-1 * (nfft - 1) / 2, (nfft - 1) / 2 + 1)) / nfft

    # cont1 = plt.contour(abs(Bspec), 4, waxis, waxis)
    cont = plt.contourf(waxis, waxis, abs(Bspec), 100, cmap=plt.cm.Spectral_r)
    plt.colorbar(cont)
    plt.title('Bispectrum estimated via the direct (FFT) method')
    plt.xlabel('f1')
    plt.ylabel('f2')
    plt.show()

    return Bspec, waxis


def test():
    qpc = sio.loadmat(here(__file__) + '/demo/qpc.mat')
    dbic = bispectrumd(qpc['zmat'], 128, 3, 64, 0)


if __name__ == '__main__':
    test()



import os
import numpy as np


def nextpow2(num):
    """
    Return the next highest power of two greater than a given value.

    Parameters
    ----------
    num : int
        Input value.

    Returns
    -------
    int
        The smallest power of two that is strictly greater than ``num``.

    Notes
    -----
    Taken from:
    https://github.com/alaiacano/frfft/blob/master/frfft.py
    """

    npow = 2
    while npow <= num:
        npow = npow * 2
    return npow


def flat_eq(x, y):
    """
    Emulate MATLAB-style linear indexing assignment.

    This function emulates the MATLAB operation ``x(:) = y``.

    Parameters
    ----------
    x : ndarray
        Input array whose shape is preserved.
    y : ndarray
        Array whose elements are assigned to ``x`` in column-major order.

    Returns
    -------
    ndarray
        An array with the same shape as ``x`` after linear assignment.
    """

    z = x.reshape(1, -1)
    z = y
    return z.reshape(x.shape)


def make_arr(arrs, axis=0):
    """
    Create arrays in a MATLAB-like concatenation manner.

    This function mimics MATLAB-style array construction by concatenating
    scalars, vectors, or matrices along a specified axis.

    Parameters
    ----------
    arrs : iterable
        Iterable of scalars, vectors, or arrays to be concatenated.
    axis : int
        Axis along which the arrays are concatenated.

    Returns
    -------
    ndarray
        A NumPy array formed by concatenating the inputs in ``arrs``.

    Examples
    --------
    .. code-block:: python

        make_arr((4, range(1, 10)), axis=0)
        array([[4],
               [1, 2, 3, 4, 5, 6, 7, 8, 9]])
    """

    a = []
    ctr = 0
    for x in arrs:
        if len(np.shape(x)) == 0:
            a.append(np.array([[x]]))
        elif len(np.shape(x)) == 1:
            a.append(np.array([x]))
        else:
            a.append(x)
        ctr += 1
    return np.concatenate(a, axis)


def shape(o, n):
    """
    Return a shape tuple padded to a given dimensionality.

    This function mimics MATLAB-style shape behavior by padding the
    shape of an array with ones if its dimensionality is less than ``n``.

    Parameters
    ----------
    o : array_like
        Input array-like object.
    n : int
        Desired number of dimensions.

    Returns
    -------
    tuple
        A tuple representing the shape of ``o`` with length ``n``.
    """

    s = o.shape
    if len(s) < n:
        x = tuple(np.ones(n - len(s)))
        return s + x
    else:
        return s


def here(f=__file__):
    """
    Return the directory containing a given file.

    By default, this returns the directory of the current script.

    Parameters
    ----------
    f : str or pathlib.Path, optional
        File path. Defaults to the current file.

    Returns
    -------
    str
        Absolute path of the directory containing ``f``.
    """

    return os.path.dirname(os.path.realpath(f))



import scipy.io as sio
from ..tools.tools import nextpow2, flat_eq, here


def cum3x(x, y, z, maxlag=0, nsamp=0, overlap=0, flag='biased', k1=0):
    """
    Estimate the third-order cross-cumulant for a fixed lag.

    Args:
        x: data vector or matrix
        y: data vector or matrix (same dimensions as x)
        z: data vector or matrix (same dimensions as x)
            if x, y, and z are matrices, columns correspond to independent
            realizations; overlap is set to 0 and nsamp is set to the
            row dimension
        maxlag: maximum lag to be computed [default = 0]
        nsamp: samples per segment [default = data length]
        overlap: percentage overlap of segments [default = 0]
            overlap is clipped to the allowed range of [0, 99]
        flag: cumulant estimation flag
            'biased': biased estimates are computed [default]
            'unbiased': unbiased estimates are computed
        k1: fixed lag in the third-order cross-cumulant C3(m, k1) [default = 0]

    Returns:
        y_cum: estimated third-order cross-cumulant sequence
            E[x*(n) y(n + m) z(n + k1)], where -maxlag <= m <= maxlag
    """

    (lx, nrecs) = x.shape
    if (lx, nrecs) != y.shape or (lx, nrecs) != z.shape:
        raise ValueError('x,y,z should have identical dimensions')

    if lx == 1:
        lx = nrecs
        nrecs = 1

    if maxlag < 0: raise ValueError('"maxlag" must be non-negative')
    if nrecs > 1: nsamp = lx
    if nsamp <= 0 or nsamp > lx: nsamp = lx
    if nrecs > 1: overlap = 0
    overlap = max(0, min(overlap, 99))

    overlap = np.fix(overlap / 100 * nsamp)
    nadvance = nsamp - overlap

    if nrecs == 1:
        nrecs = np.fix((lx - overlap) / nadvance)

    nlags = 2 * maxlag + 1
    zlag = maxlag
    y_cum = np.zeros([nlags, 1])

    if flag == 'biased':
        scale = np.ones([nlags, 1]) / nsamp
    else:
        lsamp = lx - abs(k1)
        scale = make_arr((range(lsamp - maxlag, lsamp + 1), range(lsamp - 1, lsamp - maxlag - 1, -1)), axis=1).T
        scale = np.ones([2 * maxlag + 1, 1]) / scale

    if k1 >= 0:
        indx = np.arange(nsamp - k1).T
        indz = np.arange(k1, nsamp).T
    else:
        indx = np.arange(-k1, nsamp).T
        indz = np.arange(nsamp + k1)

    ind = np.arange(nsamp)

    for k in np.arange(nrecs):
        xs = x[ind]
        xs = xs - np.mean(xs)
        ys = y[ind]
        ys = ys - np.mean(ys)
        zs = z[ind]
        zs = zs - np.mean(zs)
        zs = np.conj(zs)

        u = np.zeros([nsamp, 1])
        u[indx] = xs[indx] * zs[indz]

        y_cum[zlag] = y_cum[zlag] + np.dot(u.T, ys)

        for m in np.arange(1, maxlag + 1):
            y_cum[zlag - m] = y_cum[zlag - m] + np.dot(u[m:nsamp].T, ys[0:nsamp - m])
            y_cum[zlag + m] = y_cum[zlag + m] + np.dot(u[0:nsamp - m].T, ys[m:nsamp])

        ind = ind + int(nadvance)

    y_cum = y_cum * scale / nrecs

    return y_cum


def test():
    y = sio.loadmat(here(__file__) + '/demo/ma1.mat')['y']

    # The right results are:
    #           "biased": [0.36338   0.42762   0.77703   0.84322   0.73021  -0.13123  -0.40743]
    #           "unbiased": [0.035591   0.041841   0.075956   0.082345   0.071379  -0.012840  -0.039905]
    print(cum3x(y, y, y, 3, 100, 0, "biased"))
    print(cum3x(y, y, y, 3, 100, 0, "unbiased"))


if __name__ == '__main__':
    test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Implements Partial Directed Coherence and Direct Transfer Function
using MVAR processes.

Reference
---------
Luiz A. Baccala and Koichi Sameshima. Partial directed coherence:
a new concept in neural structure determination.
Biological Cybernetics, 84(6):463:474, 2001.
"""

# Authors: Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#
# License: BSD (3-clause)

import math
import numpy as np
from scipy import linalg, fftpack
import matplotlib.pyplot as plt


def mvar_generate(A, n, sigma, burnin=500):
    """
    Simulate a Multivariate AutoRegressive (MVAR) process.

    Args:
        A (ndarray): AR coefficients with shape (p, N, N), where N is the number of signals
                     and p is the order of the model.
        n (int): Number of time samples.
        sigma (array): Noise for each time series, shape (N,).
        burnin (int): Length of the burn-in period (in samples). Default is 500.

    Returns:
        ndarray: Time series data of shape (N, n) after burn-in.
    """
    p, N, N = A.shape
    A_2d = np.concatenate(A, axis=1)
    Y = np.zeros((n + burnin, N))

    sigma = np.diag(sigma)
    mu = np.zeros(N)

    # itération du processus
    for i in range(p, n):
        w = np.random.multivariate_normal(mu, sigma)
        Y[i] = np.dot(A_2d, Y[i - p:i][::-1, :].ravel()) + w

    return Y[burnin:].T


def cov(X, p):
    """
    Compute vector autocovariance up to order p.

    Args:
        X (ndarray): Time series data of shape (N, n).
        p (int): Order of the model.

    Returns:
        ndarray: Autocovariance up to order p with shape (p + 1, N, N).
    """
    N, n = X.shape
    R = np.zeros((p + 1, N, N))
    for k in range(p + 1):
        R[k] = (1. / float(n - k)) * np.dot(X[:, :n - k], X[:, k:].T)
    return R


def mvar_fit(X, p):
    """
    Fit an MVAR model of order p using the Yule-Walker method.

    Args:
        X (ndarray): Time series data of shape (N, n).
        p (int): Order of the model.

    Returns:
        tuple: AR coefficients (ndarray of shape (p, N, N)) and noise (array of shape (N,)).
    """
    N, n = X.shape
    gamma = cov(X, p)  # gamma(r,i,j) cov between X_i(0) et X_j(r)
    G = np.zeros((p * N, p * N))
    gamma2 = np.concatenate(gamma, axis=0)
    gamma2[:N, :N] /= 2.

    for i in range(p):
        G[N * i:, N * i:N * (i + 1)] = gamma2[:N * (p - i)]

    G = G + G.T  # big block matrix

    gamma4 = np.concatenate(gamma[1:], axis=0)

    phi = linalg.solve(G, gamma4)  # solve Yule Walker

    tmp = np.dot(gamma4[:N * p].T, phi)
    sigma = gamma[0] - tmp - tmp.T + np.dot(phi.T, np.dot(G, phi))

    phi = np.reshape(phi, (p, N, N))
    for k in range(p):
        phi[k] = phi[k].T

    return phi, sigma


def compute_order(X, p_max):
    """
    Estimate the AR order using Bayesian Information Criterion (BIC).

    Args:
        X (ndarray): Time series data of shape (N, n).
        p_max (int): Maximum model order to test.

    Returns:
        tuple: Estimated order (int) and BIC values for orders from 0 to p_max (ndarray of shape (p_max + 1,)).
    """
    N, n = X.shape

    bic = np.empty(p_max + 1)
    bic[0] = np.inf # XXX

    Y = X.T

    for p in range(1, p_max + 1):
        print (p)
        A, sigma = mvar_fit(X, p)
        A_2d = np.concatenate(A, axis=1)

        n_samples = n - p
        bic[p] = n_samples * N * math.log(2. * math.pi)
        bic[p] += n_samples * np.log(linalg.det(sigma))
        bic[p] += p * (N ** 2) * math.log(n_samples)

        sigma_inv = linalg.inv(sigma)
        S = 0.
        for i in range(p, n):
            res = Y[i] - np.dot(A_2d, Y[i - p:i][::-1, :].ravel())
            S += np.dot(res, sigma_inv.dot(res))

        bic[p] += S

    p = np.argmin(bic)
    return p, bic


def spectral_density(A, n_fft=None,sfreq=250):
    """
    Estimate Power Spectral Density (PSD) from AR coefficients.

    Args:
        A (ndarray): AR coefficients with shape (p, N, N).
        n_fft (int, optional): Length of the FFT. Defaults to None.
        sfreq (int): Sampling frequency. Default is 250.

    Returns:
        tuple: Spectral density (ndarray of shape (n_fft, N, N)) and frequencies (ndarray of shape (n_fft,)).
    """
    p, N, N = A.shape
    if n_fft is None:
        n_fft = max(int(2 ** math.ceil(np.log2(p))), 512)
    A2 = np.zeros((n_fft, N, N))
    A2[1:p + 1, :, :] = A  # start at 1 !
    fA = fftpack.fft(A2, axis=0)
    freqs = fftpack.fftfreq(n_fft,1/sfreq)
    I = np.eye(N)
    for i in range(n_fft):
        fA[i] = linalg.inv(I - fA[i])

    return fA, freqs


def dtf(A, sigma=None, n_fft=None,sfreq=250):
    """
    Compute the Direct Transfer Function (DTF).

    Args:
        A (ndarray): AR coefficients with shape (p, N, N).
        sigma (array, optional): Noise for each time series. Defaults to None.
        n_fft (int, optional): Length of the FFT. Defaults to None.
        sfreq (int): Sampling frequency. Default is 250.

    Returns:
        tuple: DTF matrix (ndarray of shape (n_fft, N, N)) and frequencies (ndarray of shape (n_fft,)).
    """
    p, N, N = A.shape

    if n_fft is None:
        n_fft = max(int(2 ** math.ceil(np.log2(p))), 512)

    H, freqs = spectral_density(A, n_fft,sfreq)
    D = np.zeros((n_fft, N, N))

    if sigma is None:
        sigma = np.ones(N)

    for i in range(n_fft):
        S = H[i]
        V = (S * sigma[None, :]).dot(S.T.conj())
        V = np.abs(np.diag(V))
        D[i] = np.abs(S * np.sqrt(sigma[None, :])) / np.sqrt(V)[:, None]

    return D, freqs


def pdc(A, sigma=None, n_fft=None,sfreq=250):
    """
    Compute Partial Directed Coherence (PDC).

    Args:
        A (ndarray): AR coefficients with shape (p, N, N).
        sigma (array, optional): Noise for each time series. Defaults to None.
        n_fft (int, optional): Length of the FFT. Defaults to None.
        sfreq (int): Sampling frequency. Default is 250.

    Returns:
        tuple: PDC matrix (ndarray of shape (n_fft, N, N)) and frequencies (ndarray of shape (n_fft,)).
    """
    p, N, N = A.shape

    if n_fft is None:
        n_fft = max(int(2 ** math.ceil(np.log2(p))), 512)

    H, freqs = spectral_density(A, n_fft,sfreq)
    P = np.zeros((n_fft, N, N))

    if sigma is None:
        sigma = np.ones(N)

    for i in range(n_fft):
        B = H[i]
        B = linalg.inv(B)
        V = np.abs(np.dot(B.T.conj(), B * (1. / sigma[:, None])))
        V = np.diag(V)  # denominator squared
        P[i] = np.abs(B * (1. / np.sqrt(sigma))[None, :]) / np.sqrt(V)[None, :]

    return P, freqs


def plot_all(freqs, P, name):
    """
    Plot a grid of subplots for visualizing the PDC or DTF matrices.

    Args:
        freqs (ndarray): Frequencies.
        P (ndarray): PDC or DTF matrix of shape (n_fft, N, N).
        name (str): Title for the plot.
    """
    m, N, N = P.shape
    pos_freqs = freqs[freqs >= 0]

    f, axes = plt.subplots(N, N)
    for i in range(N):
        for j in range(N):
            axes[i, j].fill_between(pos_freqs, P[freqs >= 0, i, j], 0)
            axes[i, j].set_xlim([0, np.max(pos_freqs)])
            axes[i, j].set_ylim([0, 1])
    plt.suptitle(name)
    plt.tight_layout()
def calculate_dtf_pdc(X,sfreq=250,kind='dtf',p=None,normalize_=True,filter_bank=None):
    """
    Calculate the Direct Transfer Function (DTF) or Partial Directed Coherence (PDC)
    from multivariate time series data.

    Args:
        X (ndarray): Time series data with shape (n_channel, n_times).
        sfreq (int): Sampling frequency in Hz. Default is 250.
        kind (str): Type of measure to compute. Options are 'dtf' for Direct Transfer Function
                    or 'pdc' for Partial Directed Coherence. Default is 'dtf'.
        p (int, optional): Order of the AR model. If None, the order will be estimated using BIC.
        normalize_ (bool): Whether to normalize the resulting matrix. Default is True.
        filter_bank (tuple, optional): Frequency range to filter. If provided, it should be a tuple
                                        (low_freq, high_freq). The output will be filtered to
                                        include only the frequencies within this range.

    Returns:
        ndarray: The computed matrix (DTF or PDC) with shape (n_channel, n_channel).
    """
    n_channel, n_times=X.shape
    if p is None:
        p = 3
    else:
        p_max = 20
        p, bic = compute_order(X, p_max=p_max)
    A_est, sigma = mvar_fit(X, p)
    sigma = np.diag(sigma)
    if kind=='dtf':
        matrix, freqs = dtf(A_est, sigma,sfreq)

    if kind == 'dpc':
        matrix, freqs = pdc(A_est, sigma,sfreq)
    if  filter_bank is not None:
        idx = np.where((freqs >= filter_bank[0]) & (freqs<= filter_bank[1]))[0]
        matrix=matrix[idx,:,:]
    matrix = np.sum(matrix, axis=0)
    if normalize_:
        for i  in range(n_channel):
            matrix[i,:]=matrix[i,:]**2/np.sum((matrix[i,:])**2)
    np.fill_diagonal(matrix, 0)
    return matrix



if __name__ == '__main__':
    plt.close('all')

    # example from the paper
    A = np.zeros((3, 5, 5))
    A[0, 0, 0] = 0.95 * math.sqrt(2)
    A[1, 0, 0] = -0.9025
    A[1, 1, 0] = 0.5
    A[2, 2, 0] = -0.4
    A[1, 3, 0] = -0.5
    A[0, 3, 3] = 0.25 * math.sqrt(2)
    A[0, 3, 4] = 0.25 * math.sqrt(2)
    A[0, 4, 3] = -0.25 * math.sqrt(2)
    A[0, 4, 4] = 0.25 * math.sqrt(2)

    # simulate processes
    n = 10 ** 4
    # sigma = np.array([0.0001, 1, 1, 1, 1])
    # sigma = np.array([0.01, 1, 1, 1, 1])
    sigma = np.array([1., 1., 1., 1., 1.])
    Y = mvar_generate(A, n, sigma)

    mu = np.mean(Y, axis=1)
    X = Y - mu[:, None]

    # estimate AR order with BIC
    if 1:
        p_max = 20
        p, bic = compute_order(X, p_max=p_max)

        # plt.figure()
        # plt.plot(np.arange(p_max + 1), bic)
        # plt.xlabel('order')
        # plt.ylabel('BIC')
        # plt.show()
    else:
        p = 3

    A_est, sigma = mvar_fit(X, p)
    sigma = np.diag(sigma)  # DTF + PDC support diagonal noise
    # sigma = None

    # compute DTF
    D, freqs = dtf(A_est, sigma)
    D=np.sum(D,axis=0)
    #plot_all(freqs, D, 'DTF')

    # compute PDC
    P, freqs = pdc(A_est, sigma)
    P = np.sum(P, axis=0)
    # plot_all(freqs, P, 'PDC')
    # plt.show()



from ..tools.tools import nextpow2, flat_eq, here


def cum2est(y, maxlag, nsamp, overlap, flag):
    """
    Estimate the second-order cumulant (covariance) function.

    This function implements CUM2EST and should be invoked via ``cumest``
    for proper parameter checking.

    Args:
        y: input data vector (column)
        maxlag: maximum lag to be computed
        samp_seg: samples per segment (<= 0 means no segmentation)
        overlap: percentage overlap of segments
        flag: covariance estimation flag
            'biased': biased estimates are computed
            'unbiased': unbiased estimates are computed

    Returns:
        y_cum: estimated covariance sequence
            C2(m), where -maxlag <= m <= maxlag
    """

    (n1, n2) = shape(y, 2)
    N = n1 * n2
    overlap = np.fix(overlap / 100 * nsamp)
    nrecord = np.fix((N - overlap) / (nsamp - overlap))
    nadvance = nsamp - overlap

    y_cum = np.zeros([maxlag + 1, 1])
    ind = np.arange(nsamp)
    y = y.ravel(order='F')

    for i in np.arange(nrecord):
        x = y[ind]
        x = x - np.mean(x)

        for k in np.arange(maxlag + 1):
            y_cum[k] = y_cum[k] + np.dot(x[0:nsamp - k].T, x[k:nsamp])

        ind = ind + int(nadvance)

    if flag == 'biased':
        y_cum = y_cum / (nsamp * nrecord)
    else:
        y_cum = y_cum / (nrecord * (nsamp - np.matrix(range(maxlag + 1)).T))
        y_cum = np.asarray(y_cum)

    if maxlag > 0:
        y_cum = make_arr([np.conj(y_cum[maxlag:0:-1]), y_cum], axis=0)

    return y_cum



from .cum3est import cum3est
from .cum4est import cum4est


def cumest(y, norder=2, maxlag=0, nsamp=None, overlap=0, flag='biased', k1=0, k2=0):
    """
    Estimate second-, third-, or fourth-order cumulants.

    This function provides a unified interface for computing cumulants of
    order 2, 3, or 4 by dispatching to the corresponding estimation routines.

    Args:
        y: input time-series (vector)
        norder: cumulant order to compute (2, 3, or 4) [default = 2]
        maxlag: maximum cumulant lag to compute [default = 0]
        nsamp: samples per segment [default = data length]
        overlap: percentage overlap of segments [default = 0]
            overlap is clipped to the allowed range of [0, 99]
        flag: cumulant estimation flag
            'biased': biased estimates are computed [default]
            'unbiased': unbiased estimates are computed
        k1: fixed lag for third- or fourth-order cumulant slices
        k2: second fixed lag for fourth-order cumulant slices

    Returns:
        y_cum: estimated cumulant sequence or slice
            C2(m), C3(m, k1), or C4(m, k1, k2), where
            -maxlag <= m <= maxlag, depending on the selected order
    """

    (ksamp, nrecs) = y.shape
    if ksamp == 1:
        ksamp = nrecs
        nrecs = 1

    if norder < 2 or norder > 4:
        raise ValueError('cumulant order must be 2, 3 or 4')

    if maxlag < 0:
        raise ValueError('"maxlag" must be non-negative')

    if nrecs > 1: nsamp = ksamp
    if nsamp <= 0 or nsamp > ksamp: nsamp = ksamp

    if nrecs > 1: overlap = 0
    overlap = max(0, min(overlap, 99))

    # estimate the cumulants
    if norder == 2:
        y_cum = cum2est(y, maxlag, nsamp, overlap, flag)
    elif norder == 3:
        y_cum = cum3est(y, maxlag, nsamp, overlap, flag, k1)
    elif norder == 4:
        y_cum = cum3est(y, maxlag, nsamp, overlap, flag, k1, k2)

    return y_cum


def test():
    y = sio.loadmat(here(__file__) + '/demo/ma1.mat')['y']

    # The right results are:
    #           "biased": [-0.12250513  0.35963613  1.00586945  0.35963613 -0.12250513]
    #           "unbiaed": [-0.12444965  0.36246791  1.00586945  0.36246791 -0.12444965]
    print(cum2est(y, 2, 128, 0, 'unbiased'))
    print(cum2est(y, 2, 128, 0, 'biased'))

    # For the 3rd cumulant:
    #           "biased": [-0.18203039  0.07751503  0.67113035  0.729953    0.07751503]
    #           "unbiased": [-0.18639911  0.07874543  0.67641484  0.74153955  0.07937539]
    print(cum3est(y, 2, 128, 0, 'biased', 1))
    print(cum3est(y, 2, 128, 0, 'unbiased', 1))

    # For testing the 4th-order cumulant
    # "biased": [-0.03642083  0.4755026   0.6352588   1.38975232  0.83791117  0.41641134 -0.97386322]
    # "unbiased": [-0.04011388  0.48736793  0.64948927  1.40734633  0.8445089   0.42303979 -0.99724968]
    print(cum4est(y, 3, 128, 0, 'biased', 1, 1))
    print(cum4est(y, 3, 128, 0, 'unbiased', 1, 1))


if __name__ == '__main__':
    test()
